(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@yellicode/elements'), require('@yellicode/templating')) :
    typeof define === 'function' && define.amd ? define(['exports', '@yellicode/elements', '@yellicode/templating'], factory) :
    (factory((global['@yellicode/typescript'] = {}),global.elements,global.templating));
}(this, (function (exports,elements,templating) { 'use strict';

    /**
     * Defines options for dealing with elements that are 'optional', meaning that they have a lower bound of 0.
     */
    (function (OptionalityModifier) {
        /**
         * Don't write any modifier.
         */
        OptionalityModifier[OptionalityModifier["Ignore"] = 0] = "Ignore";
        /**
         * Write a 'null' keyword, for example 'myProperty: string | null;';
         */
        OptionalityModifier[OptionalityModifier["NullKeyword"] = 1] = "NullKeyword";
        /**
         * Write a question token, for example 'myProperty?: string;';
         */
        OptionalityModifier[OptionalityModifier["QuestionToken"] = 2] = "QuestionToken";
    })(exports.OptionalityModifier || (exports.OptionalityModifier = {}));
    (function (EnumFeatures) {
        EnumFeatures[EnumFeatures["None"] = 0] = "None";
        EnumFeatures[EnumFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        EnumFeatures[EnumFeatures["Initializers"] = 2] = "Initializers";
        EnumFeatures[EnumFeatures["All"] = 3] = "All";
    })(exports.EnumFeatures || (exports.EnumFeatures = {}));
    (function (ClassFeatures) {
        ClassFeatures[ClassFeatures["None"] = 0] = "None";
        ClassFeatures[ClassFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        ClassFeatures[ClassFeatures["Generalizations"] = 2] = "Generalizations";
        ClassFeatures[ClassFeatures["InterfaceRealizations"] = 4] = "InterfaceRealizations";
        ClassFeatures[ClassFeatures["All"] = 7] = "All";
    })(exports.ClassFeatures || (exports.ClassFeatures = {}));
    (function (InterfaceFeatures) {
        InterfaceFeatures[InterfaceFeatures["None"] = 0] = "None";
        InterfaceFeatures[InterfaceFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        InterfaceFeatures[InterfaceFeatures["Generalizations"] = 2] = "Generalizations";
        InterfaceFeatures[InterfaceFeatures["All"] = 3] = "All";
    })(exports.InterfaceFeatures || (exports.InterfaceFeatures = {}));
    (function (PropertyFeatures) {
        PropertyFeatures[PropertyFeatures["None"] = 0] = "None";
        PropertyFeatures[PropertyFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        PropertyFeatures[PropertyFeatures["AccessModifier"] = 2] = "AccessModifier";
        PropertyFeatures[PropertyFeatures["ReadonlyModifier"] = 4] = "ReadonlyModifier";
        PropertyFeatures[PropertyFeatures["OptionalModifier"] = 8] = "OptionalModifier";
        PropertyFeatures[PropertyFeatures["Initializer"] = 16] = "Initializer";
        /**
         * Writes the so-called 'definite assignment assertion modifier' for the property if the property is required.
         * There are certain scenarios where properties can be initialized indirectly (perhaps by a helper method or dependency injection library),
         * in which case you can use the definite assignment assertion modifiers for your properties.
         */
        PropertyFeatures[PropertyFeatures["DefiniteAssignmentAssertionModifier"] = 32] = "DefiniteAssignmentAssertionModifier";
        PropertyFeatures[PropertyFeatures["All"] = 63] = "All";
        /**
         * Include all features, except the ones that are not usable on interface properties.
         */
        PropertyFeatures[PropertyFeatures["AllInterfaceProperty"] = 13] = "AllInterfaceProperty";
    })(exports.PropertyFeatures || (exports.PropertyFeatures = {}));
    (function (FunctionFeatures) {
        FunctionFeatures[FunctionFeatures["None"] = 0] = "None";
        FunctionFeatures[FunctionFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        FunctionFeatures[FunctionFeatures["AccessModifier"] = 2] = "AccessModifier";
        FunctionFeatures[FunctionFeatures["OptionalModifier"] = 4] = "OptionalModifier";
        FunctionFeatures[FunctionFeatures["All"] = 7] = "All";
    })(exports.FunctionFeatures || (exports.FunctionFeatures = {}));
    (function (ParameterFeatures) {
        ParameterFeatures[ParameterFeatures["None"] = 0] = "None";
        ParameterFeatures[ParameterFeatures["JsDocDescription"] = 1] = "JsDocDescription";
        ParameterFeatures[ParameterFeatures["OptionalModifier"] = 2] = "OptionalModifier";
        ParameterFeatures[ParameterFeatures["All"] = 3] = "All";
    })(exports.ParameterFeatures || (exports.ParameterFeatures = {}));

    class TypeScriptTypeNameProvider extends templating.DefaultTypeNameProvider {
        getTypeNameForType(type, isDataType) {
            if (!type)
                return null;
            if (isDataType) {
                if (elements.isPrimitiveBoolean(type))
                    return "boolean";
                if (elements.isPrimitiveInteger(type) || elements.isPrimitiveReal(type))
                    return "number";
                if (elements.isPrimitiveString(type))
                    return "string";
                if (elements.isPrimitiveObject(type))
                    return "any";
            }
            return super.getTypeNameForType(type, isDataType);
        }
    }

    const TYPE_BOOLEAN = 'boolean';
    const TYPE_STRING = 'string';
    const TYPE_NUMBER = 'number';
    class TypeUtility {
        static isPrimitiveType(typeName) {
            if (!typeName)
                return false;
            switch (typeName.trim()) {
                case TYPE_BOOLEAN:
                case TYPE_STRING:
                case TYPE_NUMBER:
                    return true;
                default: return false;
            }
        }
        static getPrimitiveTypeDefault(typeName) {
            if (!typeName)
                return null;
            switch (typeName.trim()) {
                case TYPE_BOOLEAN:
                    return 'false';
                case TYPE_STRING:
                    return `''`; // empty string
                case TYPE_NUMBER:
                    return '0';
                default: return null;
            }
        }
    }

    class TypeScriptModelBuilder {
        static buildClassDefinition(type, options) {
            if (!options)
                options = {};
            const definition = { name: type.name };
            const features = (options.features === undefined) ? exports.ClassFeatures.All : options.features;
            // Handle deprecated stuff
            if (options.prefix) {
                console.warn(`The 'prefix' option is deprecated. Use the 'export' or 'declare' option instead.`);
                if (options.prefix === 'export') {
                    options.export = true;
                }
                else if (options.prefix === 'declare') {
                    options.declare = true;
                }
            }
            // Description
            if ((features & exports.ClassFeatures.JsDocDescription) && type.ownedComments) {
                definition.description = TypeScriptModelBuilder.buildDescription(type);
            }
            // Generalizations
            if (elements.isClassifier(type)) {
                if (features & exports.ClassFeatures.Generalizations) {
                    definition.extends = TypeScriptModelBuilder.buildExtends(type, options.inherits);
                }
            }
            if (elements.isClass(type)) {
                // Abstract
                definition.isAbstract = type.isAbstract;
                // InterfaceRealizations
                if (features & exports.ClassFeatures.InterfaceRealizations) {
                    definition.implements = TypeScriptModelBuilder.buildImplements(type, options.implements);
                }
            }
            // Declare
            definition.declare = options.declare;
            // Export
            if (options.export == undefined) {
                definition.export = type.visibility === elements.VisibilityKind.public || type.visibility === elements.VisibilityKind.package;
            }
            else
                definition.export = options.export;
            return definition;
        }
        static buildInterfaceDefinition(type, options) {
            if (!options)
                options = {};
            const definition = { name: type.name };
            const features = (options.features === undefined) ? exports.InterfaceFeatures.All : options.features;
            // Handle deprecated stuff
            if (options.prefix) {
                console.warn(`The 'prefix' option is deprecated. Use the 'export' or 'declare' option instead.`);
                if (options.prefix === 'export') {
                    options.export = true;
                }
                else if (options.prefix === 'declare') {
                    options.declare = true;
                }
            }
            // Description
            if ((features & exports.InterfaceFeatures.JsDocDescription) && type.ownedComments) {
                definition.description = TypeScriptModelBuilder.buildDescription(type);
            }
            // Generalizations
            if (elements.isClassifier(type)) {
                if (features & exports.InterfaceFeatures.Generalizations) {
                    definition.extends = TypeScriptModelBuilder.buildExtends(type, options.inherits);
                }
            }
            // Declare
            definition.declare = options.declare;
            // Export
            if (options.export == undefined) {
                definition.export = type.visibility === elements.VisibilityKind.public || type.visibility === elements.VisibilityKind.package;
            }
            else
                definition.export = options.export;
            return definition;
        }
        static buildEnumDefinition(type, options) {
            if (!options)
                options = {};
            const definition = { name: type.name };
            const features = (options.features === undefined) ? exports.EnumFeatures.All : options.features;
            // Handle deprecated stuff
            if (options.prefix) {
                console.warn(`The 'prefix' option is deprecated. Use the 'export' or 'declare' option instead.`);
                if (options.prefix === 'export') {
                    options.export = true;
                }
                else if (options.prefix === 'declare') {
                    options.declare = true;
                }
            }
            // Description
            if ((features & exports.EnumFeatures.JsDocDescription) && type.ownedComments) {
                definition.description = TypeScriptModelBuilder.buildDescription(type);
            }
            // Declare
            definition.declare = options.declare;
            // Export
            if (options.export == undefined) {
                definition.export = type.visibility === elements.VisibilityKind.public || type.visibility === elements.VisibilityKind.package;
            }
            else
                definition.export = options.export;
            // Literals
            if (elements.isEnumeration(type) && type.ownedLiterals && type.ownedLiterals.length) {
                const members = [];
                const buildInitializers = !!(features & exports.EnumFeatures.Initializers);
                type.ownedLiterals.forEach(literal => {
                    const member = { name: literal.name };
                    if (features & exports.EnumFeatures.JsDocDescription) {
                        member.description = TypeScriptModelBuilder.buildDescription(literal);
                    }
                    if (buildInitializers && literal.specification) {
                        const specification = literal.specification;
                        member.value = elements.isLiteralInteger(specification)
                            ? specification.value
                            : specification.getStringValue();
                    }
                    members.push(member);
                });
                definition.members = members;
            }
            return definition;
        }
        static buildImplements(cls, additional) {
            const allNames = [];
            if (cls.interfaceRealizations) {
                allNames.push(...cls.interfaceRealizations.map(ir => ir.contract.name));
            }
            if (additional) {
                allNames.push(...additional);
            }
            return allNames;
        }
        static buildExtends(type, additional) {
            const allNames = [];
            if (type.generalizations) {
                allNames.push(...type.generalizations.map(g => g.general.name));
            }
            if (additional) {
                allNames.push(...additional);
            }
            return allNames;
        }
        static buildDescription(type) {
            return type.ownedComments.map(c => c.body);
        }
        static getDefaultValueString(element, typeName, defaultValue, isOptional, optionalityModifier, initializePrimitiveType, initializeArray) {
            const isPrimitive = TypeUtility.isPrimitiveType(typeName);
            const valueIfOptional = optionalityModifier == exports.OptionalityModifier.NullKeyword ? 'null' : null;
            if (element.isMultivalued()) {
                // The element is an array (of a complex type or a primitive)
                if (initializeArray) {
                    return isOptional ? valueIfOptional : '[]';
                }
                else
                    return null;
            }
            else if (isPrimitive) {
                if (defaultValue) {
                    // The type is a primitive having a default value in the model
                    return elements.isLiteralString(defaultValue) ? `'${defaultValue.value}'` : defaultValue.getStringValue();
                }
                else if (initializePrimitiveType) {
                    // The type is a primitive having no default value in the model
                    if (isOptional) {
                        return valueIfOptional;
                    }
                    else
                        return TypeUtility.getPrimitiveTypeDefault(typeName);
                }
            }
            else {
                // The element is not a primitive and no array, so it is a complex type.
                // If optionalityModifier is NullKeyword, use it.
                if (isOptional && optionalityModifier == exports.OptionalityModifier.NullKeyword) {
                    return valueIfOptional;
                }
            }
            return null;
        }
    }

    /**
     * Provides code writing functionality specific for TypeScript.
     */
    class TypeScriptWriter extends templating.CodeWriter {
        constructor(writer, options) {
            super(writer);
            if (!options)
                options = {};
            this.typeNameProvider = options.typeNameProvider || new TypeScriptTypeNameProvider();
            this.maxCommentWidth = options.maxCommentWidth || 100;
        }
        writeImports(moduleName, x) {
            if (!moduleName)
                return;
            // Ensure forward slashes in the module name
            moduleName = moduleName.replace(/\\/g, "/");
            if (x instanceof Array) {
                if (x.length === 0)
                    return;
                this.writeLine(`import { ${x.join(', ')} } from '${moduleName}';`);
                return;
            }
            // x is the alias
            if (!x) {
                x = TypeScriptWriter.makeSafeModuleName(moduleName);
            }
            this.writeLine(`import * as ${x} from '${moduleName}';`);
        }
        /**
        * Writes an indented block of decorator code, wrapped in opening and closing brackets.
        * @param contents A callback function that writes the contents.
        */
        writeDecoratorCodeBlock(decoratorName, contents) {
            this.writeLine(`@${decoratorName}({`);
            this.increaseIndent();
            if (contents)
                contents(this);
            this.decreaseIndent();
            this.writeLine('})');
        }
        /**
        * Writes an indented block of code, wrapped in opening and closing brackets.
        * @param contents A callback function that writes the contents.
        */
        writeCodeBlock(contents) {
            this.writeLine('{');
            this.increaseIndent();
            if (contents)
                contents(this);
            this.decreaseIndent();
            this.writeLine('}');
        }
        ;
        writeClassBlock(cls, contents, options) {
            if (!cls)
                return;
            let definition;
            if (elements.isType(cls)) {
                definition = TypeScriptModelBuilder.buildClassDefinition(cls, options);
            }
            else
                definition = cls;
            if (definition.description) {
                this.writeJsDocLines(definition.description);
            }
            this.writeIndent();
            if (definition.export) {
                this.write(`export `);
            }
            if (definition.declare) {
                this.write('declare ');
            }
            if (definition.isAbstract) {
                this.write('abstract ');
            }
            this.write(`class ${definition.name}`);
            if (definition.extends) {
                this.writeExtends(definition.extends);
            }
            if (definition.implements) {
                this.writeImplements(definition.implements);
            }
            // Write the contents
            this.writeEndOfLine(' {');
            this.increaseIndent();
            if (contents)
                contents(this);
            this.decreaseIndent();
            this.writeLine('}');
        }
        writeInterfaceBlock(iface, contents, options) {
            if (!iface)
                return;
            let definition;
            if (elements.isType(iface)) {
                definition = TypeScriptModelBuilder.buildInterfaceDefinition(iface, options);
            }
            else
                definition = iface;
            if (definition.description) {
                this.writeJsDocLines(definition.description);
            }
            this.writeIndent();
            if (definition.export) {
                this.write(`export `);
            }
            if (definition.declare) {
                this.write('declare ');
            }
            this.write(`interface ${definition.name}`);
            if (definition.extends) {
                this.writeExtends(definition.extends);
            }
            // Write the contents
            this.writeEndOfLine(' {');
            this.increaseIndent();
            if (contents)
                contents(this);
            this.decreaseIndent();
            this.writeLine('}');
        }
        /**
         * Writes a class or interface property.
         * @param property The property to write.
         * @param options An optional PropertyOptions object.
         */
        writeProperty(property, options) {
            if (!property)
                return;
            if (!options)
                options = {};
            const features = (options.features === undefined) ? exports.PropertyFeatures.All : options.features;
            const optionalityModifier = (options.optionality === undefined) ? exports.OptionalityModifier.QuestionToken : options.optionality;
            const makeOptional = property.isOptional() ? !!(features & exports.PropertyFeatures.OptionalModifier) : false;
            const typeName = this.getTypeName(property);
            const isOwnedByInterface = elements.isInterface(property.owner);
            const defaultValueString = ((features & exports.PropertyFeatures.Initializer) && !isOwnedByInterface) ?
                TypeScriptModelBuilder.getDefaultValueString(property, typeName, property.defaultValue, makeOptional, optionalityModifier, options.initializePrimitiveType, options.initializeArray) :
                null;
            if (features & exports.PropertyFeatures.JsDocDescription) {
                this.writeJsDocDescription(property.ownedComments);
            }
            // Start a new, indented line        
            this.writeIndent();
            // Access modifier (+ white space)
            if ((features & exports.PropertyFeatures.AccessModifier) && !isOwnedByInterface) {
                this.writeAccessModifier(property.visibility);
            }
            // Readonly modifier
            if ((features & exports.PropertyFeatures.ReadonlyModifier) && property.isReadOnly || property.isDerived) {
                this.write('readonly ');
            }
            // The name
            this.write(property.name);
            if ((optionalityModifier & exports.OptionalityModifier.QuestionToken) && makeOptional) {
                this.write('?');
            }
            if ((features & exports.PropertyFeatures.DefiniteAssignmentAssertionModifier) && !makeOptional && !defaultValueString && !isOwnedByInterface) {
                this.write('!');
            }
            // The type  
            this.write(`: ${typeName || 'any'}`);
            if (property.isMultivalued()) {
                this.write('[]');
            }
            if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
                this.write(' | null');
            }
            // Initializer if not an interface and the model has a default value
            if (defaultValueString != null) {
                this.write(` = ${defaultValueString}`);
            }
            this.writeEndOfLine(';');
        }
        writeEnumeration(enumeration, options) {
            if (!enumeration)
                return;
            let definition;
            if (elements.isType(enumeration)) {
                definition = TypeScriptModelBuilder.buildEnumDefinition(enumeration, options);
            }
            else
                definition = enumeration;
            if (definition.description) {
                this.writeJsDocLines(definition.description);
            }
            this.writeIndent();
            if (definition.export) {
                this.write(`export `);
            }
            if (definition.declare) {
                this.write('declare ');
            }
            this.write(`enum ${enumeration.name}`);
            this.writeEndOfLine();
            this.writeCodeBlock(() => {
                if (!definition.members)
                    return;
                for (let i = 0, len = definition.members.length; i < len; i++) {
                    const member = definition.members[i];
                    if (member.description) {
                        this.writeJsDocLines(member.description);
                    }
                    this.writeIndent();
                    this.write(member.name);
                    if (member.value || member.value === 0) {
                        const initialValue = (typeof member.value === "number") ?
                            member.value.toString() : `'${member.value}'`; // a string enum: wrap in quotes
                        this.write(` = ${initialValue}`);
                    }
                    if (i < len - 1) {
                        this.write(',');
                    }
                    this.writeEndOfLine();
                }
            });
        }
        /**
         * Writes a string literal type from a specified enumeration. Example: 'type Easing = 'ease-in' | 'ease-out' | 'ease-in-out';'
         * @param enumeration The enumeration.
         * @param prefix An optional prefix, such as 'export'.
         */
        writeStringLiteralType(enumeration, options) {
            if (!enumeration)
                return;
            if (!options)
                options = {};
            this.writeJsDocDescription(enumeration.ownedComments);
            this.writeIndent();
            if (options.export) {
                this.write(`export `);
            }
            if (options.declare) {
                this.write('declare ');
            }
            this.write(`type ${enumeration.name} = `);
            this.joinWrite(enumeration.ownedLiterals, ' | ', lit => `'${lit.name}'`);
            this.writeEndOfLine(';');
        }
        /**
         * Writes a function declaration without a body.
         * @param operation The operation.
         * @param options An optional FunctionOptions object.
         */
        writeFunctionDeclaration(operation, options) {
            if (!operation)
                return;
            this.writeFunctionStart(operation, options);
            this.writeEndOfLine(';');
        }
        /**
        * Writes a block of code, wrapped in an function declaration and opening and closing brackets.      *
        * @param operation The operation.
        * @param contents A callback that writes the operation contents.
        * @param options An optional FunctionOptions object.
        */
        writeFunctionBlock(operation, contents, options) {
            if (!operation)
                return;
            this.writeFunctionStart(operation, options);
            if (operation.isAbstract) {
                this.writeEndOfLine(';');
                return;
            }
            this.writeEndOfLine();
            this.writeCodeBlock((writer) => { contents(writer, operation); });
        }
        getTypeName(element) {
            if (!element)
                return null;
            if (elements.isTypedElement(element)) {
                return this.typeNameProvider ? this.typeNameProvider.getTypeName(element) : element.getTypeName();
            }
            else if (elements.isType(element)) {
                return this.typeNameProvider ? this.typeNameProvider.getTypeName(element) : element.name;
            }
            return null;
        }
        writeFunctionStart(operation, options) {
            if (!operation)
                return;
            if (!options)
                options = {};
            const features = (options.features === undefined) ? exports.FunctionFeatures.All : options.features;
            const paramFeatures = (options.parameterFeatures === undefined) ? exports.ParameterFeatures.All : options.parameterFeatures;
            // jsDoc tags 
            var jsDocLines = [];
            if (features & exports.FunctionFeatures.JsDocDescription) {
                this.pushJsDocLinesFromComments(operation.ownedComments, jsDocLines);
            }
            if (paramFeatures & exports.ParameterFeatures.JsDocDescription) {
                this.pushJsDocLinesForParameters(operation.ownedParameters, jsDocLines);
            }
            this.writeJsDocLines(jsDocLines);
            // Start a new, indented line
            this.writeIndent();
            if (!elements.isInterface(operation.owner)) {
                this.writeAccessModifier(operation.visibility);
            }
            if (operation.isStatic) {
                this.write('static ');
            }
            else if (operation.isAbstract) {
                this.write('abstract ');
            }
            this.write(operation.name);
            // If needed, make the function optional using the '?'   
            const returnParameter = operation.getReturnParameter();
            const optionalityModifier = (options.returnOptionality === undefined) ? exports.OptionalityModifier.NullKeyword : options.returnOptionality;
            const makeOptional = (returnParameter && returnParameter.isOptional()) ? !!(features & exports.FunctionFeatures.OptionalModifier) : false;
            if (makeOptional && (optionalityModifier & exports.OptionalityModifier.QuestionToken)) {
                this.write('?');
            }
            this.write('(');
            this.writeInOutParameters(operation.ownedParameters, paramFeatures, options.parameterOptionality);
            this.write('): ');
            // Write the return type        
            const returnType = this.getParameterTypeName(returnParameter) || 'void';
            this.write(returnType);
            if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
                this.write(' | null');
            }
        }
        writeInOutParameters(parameters, features, optionalityModifier) {
            if (!elements)
                return;
            if (optionalityModifier === undefined)
                optionalityModifier = exports.OptionalityModifier.NullKeyword;
            let i = 0;
            parameters.forEach((p) => {
                if (p.direction === elements.ParameterDirectionKind.return)
                    return;
                const makeOptional = p.isOptional() ? !!(features & exports.ParameterFeatures.OptionalModifier) : false;
                if (i > 0) {
                    this.write(', ');
                }
                this.write(p.name);
                if (makeOptional && (optionalityModifier & exports.OptionalityModifier.QuestionToken)) {
                    this.write('?');
                }
                const typeName = this.getParameterTypeName(p) || 'any';
                this.write(`: ${typeName}`);
                if (makeOptional && (optionalityModifier & exports.OptionalityModifier.NullKeyword)) {
                    this.write(' | null');
                }
                i++;
            });
        }
        getParameterTypeName(parameter) {
            if (!parameter)
                return null;
            var typeName = this.getTypeName(parameter) || 'any';
            if (parameter.isMultivalued()) {
                typeName = `${typeName}[]`;
            }
            return typeName;
        }
        writeAccessModifier(visibilityKind) {
            const visibilityString = TypeScriptWriter.getAccessModifierString(visibilityKind);
            if (!visibilityString)
                return;
            this.write(visibilityString);
            this.writeWhiteSpace();
        }
        writeExtends(ext) {
            if (ext.length === 0)
                return;
            this.write(' extends ');
            this.joinWrite(ext, ', ', name => name);
        }
        writeImplements(impl) {
            if (impl.length === 0)
                return;
            this.write(' implements ');
            this.joinWrite(impl, ', ', name => name);
        }
        pushJsDocLinesFromComments(comments, lines) {
            if (!comments)
                return;
            let i = 0;
            comments.forEach((p) => {
                if (!p.body) {
                    return;
                }
                if (i > 0) {
                    lines.push(''); // add a blank line
                }
                lines.push(p.body);
                i++;
            });
        }
        pushJsDocLinesForParameters(parameters, lines) {
            if (!parameters)
                return;
            parameters.forEach((p) => {
                lines.push(this.getJsDocLineForParameter(p));
            });
        }
        getJsDocLineForParameter(parameter) {
            const isReturn = parameter.direction === elements.ParameterDirectionKind.return;
            const tag = isReturn ? 'returns' : 'param';
            const commentBodies = parameter.ownedComments.map(c => c.body);
            let typeName = this.getTypeName(parameter) || 'any';
            if (parameter.isMultivalued()) {
                typeName = `${typeName}[]`;
            }
            let line = `@${tag} {${typeName}}`;
            if (!isReturn) {
                line = `${line} ${parameter.name}`;
            }
            if (commentBodies.length > 0) {
                line = `${line} ${commentBodies.join(' ')}`;
            }
            return line;
        }
        writeJsDocDescription(data) {
            const lines = [];
            if (typeof data == 'string') {
                lines.push(data);
            }
            else {
                this.pushJsDocLinesFromComments(data, lines);
            }
            this.writeJsDocLines(lines);
        }
        writeJsDocParagraph(text) {
            this.writeJsDocLines([text]);
        }
        writeJsDocLines(lines) {
            if (lines.length === 0)
                return;
            this.writeLine('/**');
            // Split here
            lines.forEach(line => {
                const lineLength = line ? line.length : 0;
                if (this.maxCommentWidth > 0 && lineLength > this.maxCommentWidth) {
                    // See if we can split the line
                    var split = templating.CodeWriterUtility.wordWrap(line, this.maxCommentWidth);
                    split.forEach(s => {
                        this.writeLine(`* ${s}`);
                    });
                }
                else
                    this.writeLine(`* ${line}`);
            });
            this.writeLine('*/');
        }
        static getAccessModifierString(visibility) {
            switch (visibility) {
                case elements.VisibilityKind.public:
                    return 'public';
                case elements.VisibilityKind.private:
                    return 'private';
                case elements.VisibilityKind.protected:
                    return 'protected';
                default:
                    return null;
            }
        }
        joinWrite(collection, separator, getStringFunc) {
            let isFirst = true;
            collection.forEach(c => {
                const value = getStringFunc(c);
                if (!value)
                    return;
                if (isFirst) {
                    isFirst = false;
                }
                else
                    this.write(separator);
                this.write(value);
            });
        }
        /**
         * Makes a TypeScript safe alias for a ES6 module name.
         */
        static makeSafeModuleName(moduleName) {
            if (moduleName.startsWith('@')) {
                moduleName = moduleName.substring(1);
            }
            const parts = moduleName.split('/');
            if (parts.length > 1) {
                // Make the module name lowerCamelCase, e.g. rename myScope/myModule to myScopeMyModule
                moduleName = templating.NameUtility.upperToLowerCamelCase(parts[0]);
                parts.forEach((p, i) => {
                    if (i > 0) {
                        moduleName += templating.NameUtility.lowerToUpperCamelCase(p);
                    }
                });
            }
            return moduleName;
        }
    }

    exports.TypeScriptTypeNameProvider = TypeScriptTypeNameProvider;
    exports.TypeScriptWriter = TypeScriptWriter;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
